import inspect
from inspect import signature
from typing import Dict, List, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel


def convert_function_call_to_json(func, *args, **kwargs):
    input_data = dict(zip(signature(func).parameters, args))
    input_data.update(kwargs)
    for key, value in input_data.items():
        if isinstance(value, BaseModel):
            input_data[key] = value.model_dump()
        elif isinstance(value, list) and all(isinstance(v, BaseModel) for v in value):
            input_data[key] = [v.model_dump() for v in value]
    return input_data


def type_to_json_schema(type_hint):
    """Convert a Python type to a JSON schema."""
    if type_hint == int:
        return {"type": "integer"}
    elif type_hint == str:
        return {"type": "string"}
    elif type_hint == float:
        return {"type": "number"}
    elif type_hint == bool:
        return {"type": "boolean"}
    elif inspect.isclass(type_hint) and issubclass(type_hint, BaseModel):
        return type_hint.model_json_schema()
    elif hasattr(type_hint, "__origin__"):
        # Handling generic types (List, Dict, etc.)
        if get_origin(type_hint) == Union and type(None) in get_args(type_hint):
            return type_to_json_schema(get_args(type_hint)[0])  # for Optional
        elif type_hint.__origin__ in (List, list):
            item_type = get_args(type_hint)[0]
            if inspect.isclass(item_type):
                return {"type": "array", "items": type_to_json_schema(item_type)}
            if get_origin(item_type) == List:
                return {"type": "array", "items": type_to_json_schema(item_type)}
            elif inspect.isclass(item_type) and issubclass(item_type, BaseModel):
                return {"type": "array", "items": item_type.schema()}
            else:
                return {"type": "array", "items": type_to_json_schema(item_type)}
        elif get_origin(type_hint) == Dict:
            return {"type": "object"}
        else:
            raise NotImplementedError(
                f"Type {type_hint} is not supported for JSON schema generation.{get_origin(type_hint)}"
            )
    else:
        return {"type": "object"}


def get_input_schema(func):
    """Generate input JSON schema from a function signature."""
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)

    schema = {"type": "object", "properties": {}, "required": []}

    for param in sig.parameters.values():
        param_type = type_hints.get(
            param.name, str
        )  # Default to string if type is not specified
        schema["properties"][param.name] = type_to_json_schema(param_type)
        if param.default is inspect.Parameter.empty:
            schema["required"].append(param.name)

    if not schema["properties"]:
        return {
            "type": "object",
            "description": "No input parameters",
            "properties": {},
            "additionalProperties": False,
        }

    return schema


def get_output_schema(func):
    """Generate output JSON schema from a function signature."""
    return_type = get_type_hints(func).get("return", None)
    if return_type:
        schema = type_to_json_schema(return_type)
    else:
        schema = {"type": "unknown"}

    return schema



from unittest.mock import MagicMock, patch

from opperai import Client
from opperai.types.indexes import DocumentIn


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_retrieve(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: [
            {
                "score": 2.3,
                "id": 1,
                "content": "Bonjour",
                "metadata": {"source": "test"},
            }
        ],
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    resp = client.indexes.retrieve(index_id=1, query="test", k=1)

    assert resp[0].content == "Bonjour"
    assert resp[0].metadata == {"source": "test"}
    mock_do_request.assert_called_once()


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_create(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: {"dataset_id": 123},
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    dataset_id = client.indexes.create(name="test_index")

    assert dataset_id == 123
    mock_do_request.assert_called_once_with(
        "POST",
        "/v1/indexes",
        json={"name": "test_index"},
    )


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_index(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: {
            "id": 1,
            "uuid": "506768ec-54ac-4233-8a7b-ea9d1f12b879",
            "key": "test_key",
        },
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    doc_in = DocumentIn(content="Hello", metadata={"source": "test"})
    doc_out = client.indexes.index(doc=doc_in)

    assert doc_out.id == 1
    assert str(doc_out.uuid) == "506768ec-54ac-4233-8a7b-ea9d1f12b879"
    assert doc_out.key == "test_key"
    mock_do_request.assert_called_once()


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_list(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: [
            {"id": 1, "name": "test_index_1", "created_at": "2021-07-21T17:32:28Z"},
            {"id": 2, "name": "test_index_2", "created_at": "2021-07-22T17:32:28Z"},
        ],
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    indexes = client.indexes.list()

    assert len(indexes) == 2
    assert indexes[0].id == 1
    assert indexes[0].name == "test_index_1"
    assert indexes[1].id == 2
    assert indexes[1].name == "test_index_2"
    mock_do_request.assert_called_once_with(
        "GET",
        "/v1/indexes",
    )



import asyncio
import inspect
import json
from functools import wraps
from typing import List, get_args, get_origin, get_type_hints

from pydantic import BaseModel

from opperai import AsyncClient, Client
from opperai.types import ChatPayload, FunctionDescription, Message

from ._schemas import convert_function_call_to_json, get_output_schema


def fn(path=None, client=None, json_encoder=None):
    def decorator(func):
        if isinstance(client, AsyncClient):
            sync_client = Client(api_key=client.api_key, api_url=client.api_url)
        elif isinstance(client, Client):
            sync_client = client
        else:
            sync_client = Client()
        func_path = path or func.__name__

        function = FunctionDescription(
            path=func_path,
            description=func.__doc__,
            instructions=f"Operation: {func.__name__}\n\nOperation description: {func.__doc__}",
            out_schema=get_output_schema(func),
        )
        sync_client.functions.create_function(function)

        if asyncio.iscoroutinefunction(func):
            c = AsyncClient() or client
        else:
            c = sync_client or client

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            input = convert_function_call_to_json(func, *args, **kwargs)
            payload = ChatPayload(
                messages=[
                    Message(role="user", content=json.dumps(input, cls=json_encoder))
                ]
            )
            response = await c.functions.chat(func_path, payload)
            answer = response.json_payload

            return_type = get_type_hints(func).get("return")

            if return_type is not None:
                if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                    answer = return_type(**answer)
                elif (
                    (get_origin(return_type) == list or get_origin(return_type) == List)
                    and inspect.isclass(get_args(return_type)[0])
                    and issubclass(get_args(return_type)[0], BaseModel)
                ):
                    answer = [get_args(return_type)[0](**item) for item in answer]
            return answer

        def sync_wrapper(*args, **kwargs):
            input = convert_function_call_to_json(func, *args, **kwargs)
            payload = ChatPayload(
                messages=[
                    Message(role="user", content=json.dumps(input, cls=json_encoder))
                ]
            )
            answer = c.functions.chat(func_path, payload).json_payload

            return_type = get_type_hints(func).get("return")
            if return_type is not None:
                if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                    answer = return_type(**answer)
                elif (
                    (get_origin(return_type) == list or get_origin(return_type) == List)
                    and inspect.isclass(get_args(return_type)[0])
                    and issubclass(get_args(return_type)[0], BaseModel)
                ):
                    answer = [get_args(return_type)[0](**item) for item in answer]
            return answer

        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper

    return decorator



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field


class Message(BaseModel):
    role: str
    content: str


class ChatPayload(BaseModel):
    messages: List[Message]

    class Config:
        extra = "allow"


class FileMetadata(BaseModel):
    file_name: str


class ContextData(BaseModel):
    index_id: int
    content: str
    metadata: Union[Optional[dict], FileMetadata] = Field(default=None)


class StreamingChunk(BaseModel):
    delta: Optional[str] = None
    error: Optional[str] = None
    context: Optional[List[ContextData]] = None


class FunctionResponse(BaseModel):
    message: Optional[str] = None
    json_payload: Optional[Union[dict, List, Any]] = None
    error: Optional[str] = None
    context: Optional[List[ContextData]] = None


class FunctionDescription(BaseModel):
    id: Optional[int] = None
    path: str = Field(
        ...,
        pattern=r"^[a-zA-Z0-9_]+(\/[a-zA-Z0-9_-]+)*$",
        description="Path should not contain characters that could break URLs.",
    )
    description: str
    input_schema: Optional[Dict[str, Any]] = None
    out_schema: Optional[Dict[str, Any]] = None
    instructions: str
    dataset_ids: Optional[List[int]] = None



from opperai._http_clients import _async_http_client
from opperai.types import (
    ChatPayload,
    FunctionDescription,
    FunctionResponse,
    StreamingChunk,
)
from opperai.types.exceptions import APIError


class AsyncFunctions:
    def __init__(self, http_client: _async_http_client):
        self.http_client = http_client

    async def _create_function(self, function: FunctionDescription) -> int:
        response = await self.http_client.do_request(
            "POST",
            "/api/v1/functions",
            json=function.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to create function {function.path} with status {response.status_code}"
            )
        return response.json()["id"]

    async def update_function(self, function: FunctionDescription) -> int:
        response = await self.http_client.do_request(
            "POST",
            f"/api/v1/functions/{function.id}",
            json=function.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to update function {function.path} with status {response.status_code}"
            )
        return response.json()["id"]

    async def get_function_by_path(self, function_path: str) -> FunctionDescription:
        response = await self.http_client.do_request(
            "GET",
            f"/api/v1/functions/by_path/{function_path}",
        )
        if response.status_code == 404:
            return None
        if response.status_code != 200:
            raise APIError(
                f"Failed to get function {function_path} with status {response.status_code}"
            )

        return FunctionDescription(**response.json())

    async def get_function_by_id(self, function_id: str) -> FunctionDescription:
        response = await self.http_client.do_request(
            "GET",
            f"/api/v1/functions/{function_id}",
        )
        if response.status_code == 404:
            return None
        if response.status_code != 200:
            raise APIError(
                f"Failed to get function {function_id} with status {response.status_code}"
            )

        return FunctionDescription(**response.json())

    async def create_function(
        self, function: FunctionDescription, update: bool = True
    ) -> int:
        f = await self.get_function_by_path(function.path)
        if f is None:
            return await self._create_function(function)
        elif update:
            function.id = f.id
            return await self.update_function(function)
        return None

    async def chat(
        self, function_path, data: ChatPayload, stream=False
    ) -> FunctionResponse:
        if stream:
            return self._chat_stream(function_path, data)
        serialized_data = data.model_dump()

        response = await self.http_client.do_request(
            "POST",
            f"/v1/chat/{function_path}",
            json=serialized_data,
            params={"stream": "true"} if stream else None,
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to run function {function_path} with status {response.status_code}"
            )
        return FunctionResponse(**response.json())

    async def _chat_stream(
        self,
        function_path,
        data: ChatPayload,
    ) -> FunctionResponse:
        gen = self.http_client.stream(
            "POST", f"/v1/chat/{function_path}?stream=True", json=data.model_dump()
        )
        async for item in gen:
            yield StreamingChunk(**item)



from opperai._http_clients import _http_client
from opperai.types.exceptions import APIError
from opperai.types.indexes import (
    DocumentIn,
    DocumentOut,
    IndexOut,
    IndexRetrieveResponse,
)


class Indexes:
    def __init__(self, http_client: _http_client):
        self.http_client = http_client

    def create(self, name: str) -> int:
        response = self.http_client.do_request(
            "POST",
            "/v1/indexes",
            json={"name": name},
        )
        if response.status_code != 200:
            raise APIError(f"Failed to create index with status {response.status_code}")
        return response.json()["dataset_id"]

    def delete(self, index_id: int):
        response = self.http_client.do_request(
            "DELETE",
            f"/v1/indexes/{index_id}",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to delete index with status {response.status_code}")
        return response.json()

    def list(self):
        response = self.http_client.do_request(
            "GET",
            "/v1/indexes",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to list indexes with status {response.status_code}")
        return [IndexOut.model_validate(item) for item in response.json()]

    def index(self, doc: DocumentIn) -> DocumentOut:
        response = self.http_client.do_request(
            "POST",
            "/v1/indexes/index",
            json=doc.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(f"Failed to add document with status {response.status_code}")
        return DocumentOut.model_validate(response.json())

    def retrieve(self, index_id: int, query: str, k: int):
        response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{index_id}/query",
            json={"q": query, "k": k},
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to retrieve index {index_id} with status {response.status_code}"
            )
        return [IndexRetrieveResponse.model_validate(item) for item in response.json()]



import datetime
from typing import Optional

from pydantic import UUID4, BaseModel, Field


class IndexRetrieveResponse(BaseModel):
    content: str
    metadata: dict


class DocumentIn(BaseModel):
    id: Optional[str] = None
    key: Optional[str] = None
    content: str = Field(..., min_length=1)
    metadata: dict = dict()


class DocumentOut(BaseModel):
    id: int
    uuid: UUID4
    key: str


class IndexOut(BaseModel):
    id: int
    name: str
    created_at: datetime.datetime
